/* AsmBuilder.hpp */

#ifndef ASMBUILDER_HPP
#define ASMBUILDER_HPP

#include <vector>
#include <map>
#include <string>
#include <cstdint>
#include <initializer_list>
#include <sstream>

enum Register {
	// Abstract, might be implemented to refer to different registers on different platforms.
	// x86 order is purposely not alphabetical to match the ISA

	REG__INVALID,

	// x86_32
	REG_A,  // eax	
	REG_C,  // ecx
	REG_D,  // edx
	REG_B,  // ebx		
	REG_SP, // esp
	REG_BP, // ebp
	REG_SI, // esi
	REG_DI, // edi

};

enum BackfillableReferenceTypes {
	BRT__INVALID,

	BRT_RELATIVE_32
};

struct BackfillableReference {
	uint32_t datapos;
	enum BackfillableReferenceTypes type;
	std::string ref_target;
};

struct Section {
	std::string data;
	std::vector<BackfillableReference> MissingReferences;
};

struct LabelTarget {
	Section* section;
	size_t offset;
};

class AsmBuilder {

protected:

	std::map<std::string, Section> sections;
	std::map<std::string, LabelTarget> labels;
	Section* curSection = 0;

public:

	std::ostringstream errors;

	virtual ~AsmBuilder() {};

	/* */

	void changeSection(std::string name);

	uint32_t getCurrentSectionOffset();

	/* */

	void raw8s(std::initializer_list<uint8_t> vals);

	virtual void raw8(uint8_t val) = 0;

	virtual void raw16(uint16_t val) = 0;

	virtual void raw32(uint32_t val) = 0;

	virtual void rawData(const char* str, size_t len) = 0;

	virtual void rawData(std::string str) = 0;

	/* */

	virtual void pushReg(Register r) = 0;

	virtual void popReg(Register r) = 0;

	virtual void pushRegOffset(Register r, int32_t offset) = 0;

	virtual void movRegReg(Register dest, Register src) = 0;

	virtual void movRegImm32(Register dest, uint32_t val) = 0;

	virtual void movRegRegOffset(Register dest, Register srcptr, int32_t srcoffset) = 0;

	virtual void movRegOffsetReg(Register destptr, int32_t destoffset, Register src) = 0;

	virtual void addRegReg(Register accumulator, Register src) = 0;

	virtual void addRegImm32(Register accumulator, uint32_t val) = 0;

	virtual void clearRegister(Register r) = 0;

	/* */

	virtual void exportGlobalHere(std::string name) = 0;

	virtual void label(std::string name) = 0;

	virtual void jumpUnconditional(std::string label) = 0;

	virtual void jumpIfLesserOrEqual(std::string label) = 0;

	virtual void jumpIfLesser(std::string label) = 0;

	virtual void jumpIfEqual(std::string label) = 0;

	virtual void jumpIfNotEqual(std::string label) = 0;

	virtual void jumpIfGreater(std::string label) = 0;

	virtual void jumpIfGreaterOrEqual(std::string label) = 0;

	virtual void noop() = 0;

	/* */

	virtual void callByLabel(std::string label) = 0;

	virtual void callFromRegister(Register r) = 0;

	virtual void leave() = 0;

	virtual void returnFunc() = 0;

	virtual void returnFunc(uint16_t stacksize) = 0;

	/* */

	virtual bool renderObject(std::string& ret) = 0;

};

#endif 
