/* AsmBuilder_x8632obj.cpp */

#include "AsmBuilder_x8632obj.hpp"
#include <cstdint>
#include <algorithm>

typedef uint32_t uint32;
typedef uint16_t uint16;
typedef int16_t sint16;
typedef uint8_t uchar;

struct coffHeader{
	uint16 machine;
	uint16 numSections;
	uint32 timeStamp;
	uint32 ptrSymbolTable; /* zero/deprecated */
	uint32 numSymbols;     /* zero/deprecated */
	uint16 sizeOfOptionalHeader;
	uint16 characteristics;
};

struct sectionHeader{
	char name[8];
	uint32 virtualSize;
	uint32 virtualAddress;
	uint32 sizeOfRawData;
	uint32 fpRawData;
	uint32 fpRelocations;
	uint32 fpLineNumbers; /* zero, deprecated */
	uint16 numRelocations;
	uint16 numLineNumbers; /* zero, deprecated */
	uint32 characteristics;
};

struct relocation{
	uint32 virtualAddress;
	uint32 symbolTableIndex;
	uint16 type;
};

#define SIZEOF_RELOCATION 10

struct coffSymbol{
	union {
		char shortName[8];
		struct {
			uint32 zero;
			uint32 stringTableOffset;
		} offset;
	} name;

	uint32 value;
	sint16 sectionNumber; /* signed */
	uint16 type;
	uchar storageClass;
	uchar numAuxSymbols;
};

#define MACHINE_TYPE_i386 0x14c

#define IMAGE_REL_I386_REL32 0x0014

/* ###### */

void AsmBuilder_x8632obj::exportGlobalHere(std::string name) {
	LabelTarget l;
	l.offset = getCurrentSectionOffset();
	l.section = curSection;
	globalExports[name] = l;
}

bool AsmBuilder_x8632obj::renderObject(std::string& ret) {

	patchReferences();

	/*
		At this point, we are expecting that references are all joined wherever possible and all 
		the remaining necessary relocations will be in the MissingReferences array
	*/

	/* STRING TABLE */
	std::string stringTable;
	stringTable.append("\x00\x00\x00\x00", 4); // first four bytes are the length

	/* SYMBOL TABLE */
	std::string symbolTable;
	uint32_t nextSymbolTableIndex = 0;

	/* IMAGE_FILE_HEADER */
	coffHeader ch = { 0 };
	ch.machine = MACHINE_TYPE_i386;
	ch.numSections = sections.size();
	ch.sizeOfOptionalHeader = 0;
	ch.characteristics = 0x0104; /* 32BIT_MACHINE | LINE_NUMS_STRIPPED */ 
	size_t filePtr = sizeof(coffHeader)+(sections.size() * sizeof(sectionHeader));

	size_t numSymbols = 0;

	size_t filePtrAfterData = filePtr;
	for (auto section : sections) {
		filePtrAfterData += section.second.data.size();
		numSymbols += section.second.MissingReferences.size();
	}

	size_t filePtrNextRelocationSection = filePtrAfterData;

	size_t filePtrSymbolTable = filePtrNextRelocationSection + (numSymbols * sizeof(relocation));

	ch.ptrSymbolTable = filePtrSymbolTable;

	ret.append((const char*)&ch, sizeof(coffHeader));

	//

	for (auto section : sections) {
		
		/* IMAGE_SECTION_HEADER */
		sectionHeader sh = { 0 };
		
		std::string tmp = "/" + std::to_string(stringTable.size());
		memcpy(sh.name, tmp.c_str(), std::min(tmp.length(), sizeof(sh.name)));

		stringTable.append(section.first);
		stringTable.append("\x00", 1);

		if (section.second.data.length()) {
			
			sh.virtualSize = section.second.data.length();
			sh.sizeOfRawData = section.second.data.length();
			sh.fpRawData = filePtr;

			filePtr += section.second.data.length();

		}

		if (section.second.MissingReferences.size()) {

			sh.numRelocations = section.second.MissingReferences.size();
			sh.fpRelocations = filePtrNextRelocationSection;

			filePtrNextRelocationSection += (section.second.MissingReferences.size() * SIZEOF_RELOCATION);

		}

		// Characteristics {{{ 

		sh.characteristics |= 0x00300000; /* IMAGE_SCN_ALIGN_4BYTES */
		sh.characteristics |= 0x40000000; /* IMAGE_SCN_MEM_READ */

		if (section.first.find(".text") != std::string::npos) {
			sh.characteristics |= 0x20; /* IMAGE_SCN_CNT_CODE */
			sh.characteristics |= 0x20000000; /* IMAGE_SCN_MEM_EXECUTE */
			
		} else if (section.first.find(".rdata")) {
			sh.characteristics |= 0x40; /* IMAGE_SCN_CNT_INITIALIZED_DATA */

		} else { /* assume r/w */
			sh.characteristics |= 0x80000000; /* IMAGE_SCN_MEM_WRITE */

		}

		// }}} 

		ret.append((char*)&sh, sizeof(sectionHeader));

	}

	for (auto section : sections) {

		/* SECTION DATA */
		ret.append(section.second.data.c_str(), section.second.data.length());

	}

	/* IMAGE_RELOCATION */

	for (auto section : sections) {
		for (auto mref : section.second.MissingReferences) {
			
			relocation rc = { 0 };
			if (mref.type == BackfillableReferenceTypes::BRT_RELATIVE_32) {
				rc.type = IMAGE_REL_I386_REL32;

			} else {
				errors << "FATAL: Unexpected BackfillableReferenceType( " << std::to_string(rc.type) << "\"\n";
				return false;
			}

			rc.virtualAddress = mref.datapos; // easy 

			// Create symbol for symbol table

			coffSymbol sym = { 0 };
			
			sym.name.offset.stringTableOffset = stringTable.size();
			stringTable.append(mref.ref_target);
			stringTable.append("\x00", 1);
			sym.storageClass = 0x02; /* IMAGE_SYM_CLASS_EXTERNAL */
			
			symbolTable.append(static_cast<char*>(static_cast<void*>(&sym)), sizeof(coffSymbol));

			rc.symbolTableIndex = nextSymbolTableIndex++;

			ret.append((char*)&rc, sizeof(relocation));

		}
	}

	/* IMAGE_SYMBOL Table */

	ret += symbolTable;

	/* IMAGE_SYMBOL String Table */

	ret += stringTable;

	/* Done */

	return true;
}
