/* AsmBuilder_x8632raw.cpp */

#include "AsmBuilder_x8632raw.hpp"

void AsmBuilder_x8632raw::raw8(uint8_t val) {
	curSection->data.push_back(val);
}

void AsmBuilder_x8632raw::raw16(uint16_t val) {
	uint8_t msb = (val >> 8) & 0xFF;
	uint8_t lsb = val & 0xFF;
	curSection->data.push_back(lsb);
	curSection->data.push_back(msb);
}

void AsmBuilder_x8632raw::raw32(uint32_t val) {
	curSection->data.append((char*)( &val ), sizeof(uint32_t));
}

void AsmBuilder_x8632raw::rawData(const char* str, size_t len) {
	curSection->data.append(str, len);
}

void AsmBuilder_x8632raw::rawData(std::string str) {
	curSection->data.append(str);
}

/* */

void AsmBuilder_x8632raw::pushReg(Register r) {
	switch (r) {
		case REG_A:		return raw8(0x50);
		case REG_C:		return raw8(0x51);
		case REG_D:		return raw8(0x52);
		case REG_B:		return raw8(0x53);
		case REG_SP:	return raw8(0x54);
		case REG_BP:	return raw8(0x55);
		case REG_SI:	return raw8(0x56);
		case REG_DI:	return raw8(0x57);
	}
}

void AsmBuilder_x8632raw::popReg(Register r) {
	switch (r) {
		case REG_A:		return raw8(0x58);
		case REG_C:		return raw8(0x59);
		case REG_D:		return raw8(0x5a);
		case REG_B:		return raw8(0x5b);
		case REG_SP:	return raw8(0x5c);
		case REG_BP:	return raw8(0x5d);
		case REG_SI:	return raw8(0x5e);
		case REG_DI:	return raw8(0x5f);
	}
}

void AsmBuilder_x8632raw::pushRegOffset(Register r, int32_t offset) {
	// push dword[ebp+8]

	if (r != REG_BP && offset == 0) {
		// zero-byte suffix (doesn't exist for EBP)

		switch (r) {
			case REG_A:		return raw8s({ 0xff, 0x30 });
			case REG_C:		return raw8s({ 0xff, 0x31 });
			case REG_D:		return raw8s({ 0xff, 0x32 });
			case REG_B:		return raw8s({ 0xff, 0x33 });
			case REG_SP:	return raw8s({ 0xff, 0x34, 0x24 });
			case REG_SI:	return raw8s({ 0xff, 0x36 });
			case REG_DI:	return raw8s({ 0xff, 0x37 });
		}

	} else if (offset >= -127 && offset <= 127) {
		// one-byte suffix

		uint8_t offset8 = (offset & 0xFF); // cast

		switch (r) {
			case REG_A:		return raw8s({ 0xff, 0x70, offset8 });
			case REG_C:		return raw8s({ 0xff, 0x71, offset8 });
			case REG_D:		return raw8s({ 0xff, 0x72, offset8 });
			case REG_B:		return raw8s({ 0xff, 0x73, offset8 });
			case REG_SP:	return raw8s({ 0xff, 0x74, 0x24, offset8 });
			case REG_BP:	return raw8s({ 0xff, 0x75, offset8 });
			case REG_SI:	return raw8s({ 0xff, 0x76, offset8 });
			case REG_DI:	return raw8s({ 0xff, 0x77, offset8 });
		}

	} else {
		// four-byte suffix

		switch (r) {
			case REG_A:		raw8s({ 0xff, 0xb0 }); break;
			case REG_C:		raw8s({ 0xff, 0xb1 }); break;
			case REG_D:		raw8s({ 0xff, 0xb2 }); break;
			case REG_B:		raw8s({ 0xff, 0xb3 }); break;
			case REG_SP:	raw8s({ 0xff, 0xb4, 0x24 }); break;
			case REG_BP:	raw8s({ 0xff, 0xb5 }); break;
			case REG_SI:	raw8s({ 0xff, 0xb6 }); break;
			case REG_DI:	raw8s({ 0xff, 0xb7 }); break;
		}
		raw32(offset);

	}
}

void AsmBuilder_x8632raw::movRegImm32(Register dest, uint32_t val) {
	switch (dest) {
		case REG_A:		raw8(0xb8); break;
		case REG_C:		raw8(0xb9); break;
		case REG_D:		raw8(0xba); break;
		case REG_B:		raw8(0xbb); break;
		case REG_SP:	raw8(0xbc); break;
		case REG_BP:	raw8(0xbd); break;
		case REG_SI:	raw8(0xbe); break;
		case REG_DI:	raw8(0xbf); break;
	}
	raw32(val);
}

uint8_t combineRegistersByStep(Register step1, Register step8) {

	uint8_t offset = 0;

	switch (step1) {
		case REG_A:		offset = 0; break;
		case REG_C:		offset = 1; break;
		case REG_D:		offset = 2; break;
		case REG_B:		offset = 3; break;
		case REG_SP:	offset = 4; break;
		case REG_BP:	offset = 5; break;
		case REG_SI:	offset = 6; break;
		case REG_DI:	offset = 7; break;
	}

	switch (step8) {
		case REG_A:		return 0x00 + offset;
		case REG_C:		return 0x08 + offset;
		case REG_D:		return 0x10 + offset;
		case REG_B:		return 0x18 + offset;
		case REG_SP:	return 0x20 + offset;
		case REG_BP:	return 0x28 + offset;
		case REG_SI:	return 0x30 + offset;
		case REG_DI:	return 0x38 + offset;
	}

}

void AsmBuilder_x8632raw::movRegReg(Register dest, Register src) {
	raw8(0x89);
	raw8(0xc0 + combineRegistersByStep(dest, src));
}

void AsmBuilder_x8632raw::movRegRegOffset(Register dest, Register srcptr, int32_t srcoffset) {
	// mov reg, dword[eax + y]

	if (srcptr != REG_BP && srcoffset == 0) {
		// zero-byte version (doesn't exist for ebp)

		raw8(0x03);
		raw8(0x00 + combineRegistersByStep(srcptr, dest));
		if (srcptr == REG_SP) {
			raw8(0x24);
		}

	} else if (srcoffset >= -127 && srcoffset <= 127) {
		// one-byte version

		raw8(0x03);
		raw8(0x40 + combineRegistersByStep(srcptr, dest));
		if (srcptr == REG_SP) {
			raw8(0x24);
		}

		uint8_t srcoffset8 = (srcoffset & 0xFF);
		raw8(srcoffset8);

	} else {
		// four-byte version

		raw8(0x03);
		raw8(0x80 + combineRegistersByStep(srcptr, dest));
		if (srcptr == REG_SP) {
			raw8(0x24);
		}

		raw32(srcoffset);

	}

}

void AsmBuilder_x8632raw::movRegOffsetReg(Register destptr, int32_t destoffset, Register src) {
	// mov dword[eax + y], reg

	if (src != REG_BP && destoffset == 0) {
		// zero-byte version (doesn't exist for ebp)

		raw8(0x01);
		raw8(0x00 + combineRegistersByStep(destptr, src));
		if (destptr == REG_SP) {
			raw8(0x24);
		}

	} else if (destoffset >= -127 && destoffset <= 127) {
		// one-byte version

		raw8(0x01);
		raw8(0x40 + combineRegistersByStep(destptr, src));
		if (destptr == REG_SP) {
			raw8(0x24);
		}

		uint8_t destoffset8 = (destoffset & 0xFF);
		raw8(destoffset8);

	} else {
		// four-byte version

		raw8(0x01);
		raw8(0x80 + combineRegistersByStep(destptr, src));
		if (destptr == REG_SP) {
			raw8(0x24);
		}

		raw32(destoffset);

	}
}

void AsmBuilder_x8632raw::addRegReg(Register accumulator, Register src) {
	raw8(0x01);
	raw8(0xc0 + combineRegistersByStep(accumulator, src));
}

void AsmBuilder_x8632raw::addRegImm32(Register accumulator, uint32_t val) {
	if (val >= -127 && val <= 127) {
		// one-byte version
		uint8_t val8 = (val & 0xFF);

		switch (accumulator) {
			case REG_A:		raw8s({ 0x83, 0xc0 }); break;
			case REG_C:		raw8s({ 0x83, 0xc1 }); break;
			case REG_D:		raw8s({ 0x83, 0xc2 }); break;
			case REG_B:		raw8s({ 0x83, 0xc3 }); break;
			case REG_SP:	raw8s({ 0x83, 0xc4 }); break;
			case REG_BP:	raw8s({ 0x83, 0xc5 }); break;
			case REG_SI:	raw8s({ 0x83, 0xc6 }); break;
			case REG_DI:	raw8s({ 0x83, 0xc7 }); break;
		}
		raw8(val8);

	} else {
		// four-byte version

		switch (accumulator) {
			case REG_A:		raw8s({ 0x05 }); break; // shorter
			case REG_C:		raw8s({ 0x81, 0xc1 }); break;
			case REG_D:		raw8s({ 0x81, 0xc2 }); break;
			case REG_B:		raw8s({ 0x81, 0xc3 }); break;
			case REG_SP:	raw8s({ 0x81, 0xc4 }); break;
			case REG_BP:	raw8s({ 0x81, 0xc5 }); break;
			case REG_SI:	raw8s({ 0x81, 0xc6 }); break;
			case REG_DI:	raw8s({ 0x81, 0xc7 }); break;
		}
		raw32(val);

	}
}

void AsmBuilder_x8632raw::clearRegister(Register r) {
	switch (r) {
		case REG_A:		return raw8s({ 0x31, 0xc0 });
		case REG_C:		return raw8s({ 0x31, 0xc9 });
		case REG_D:		return raw8s({ 0x31, 0xd2 });
		case REG_B:		return raw8s({ 0x31, 0xdb });
		case REG_SP:	return raw8s({ 0x31, 0xe4 });
		case REG_BP:	return raw8s({ 0x31, 0xed });
		case REG_SI:	return raw8s({ 0x31, 0xf6 });
		case REG_DI:	return raw8s({ 0x31, 0xff });
	}
}

/* */

void AsmBuilder_x8632raw::label(std::string name) {
	LabelTarget l;
	l.offset = getCurrentSectionOffset();
	l.section = curSection;
	labels[name] = l;
}

void AsmBuilder_x8632raw::exportGlobalHere(std::string name) {
	// no-op
}

void AsmBuilder_x8632raw::jumpWithShortVersion(std::string label, std::initializer_list<uint8_t> shortver, std::initializer_list<uint8_t> longver) {
	// There is a version of this opcode with a rel8 target instead of a rel32, which
	//   would allow us to produce tighter output. Fill in backward jumps immediately
	//   using this short version if possible - although there's not much we can do
	//   about forward jumps at present

	uint32_t cur_offset = getCurrentSectionOffset(); // not counting shortver.length (problem?)
	if (labels.find(label) != labels.end()) {
		int32_t rel_offset = labels[label].offset - cur_offset;
		if (rel_offset >= -127 && rel_offset < 0) {

			// Use short version
			uint8_t rel_offset8 = (rel_offset & 0xFF);
			raw8s(shortver);
			raw8( rel_offset8 );

			return;
		}
	}

	// Couldn't use short version

	raw8s(longver);
	curSection->MissingReferences.emplace_back(
		BackfillableReference{ getCurrentSectionOffset(), BRT_RELATIVE_32, label }
	);
	raw32(0x00000000);
}

void AsmBuilder_x8632raw::jumpUnconditional(std::string label) {
	// jmp [target]

	jumpWithShortVersion(label, { 0xeb }, { 0xe9 });
}

void AsmBuilder_x8632raw::jumpIfLesserOrEqual(std::string label) {
	jumpWithShortVersion(label, { 0x7e }, { 0x0f, 0x8e });
}

void AsmBuilder_x8632raw::jumpIfLesser(std::string label) {
	jumpWithShortVersion(label, { 0x7c }, { 0x0f, 0x8c });
}

void AsmBuilder_x8632raw::jumpIfEqual(std::string label) {
	jumpWithShortVersion(label, { 0x74 }, { 0x0f, 0x84 });
}

void AsmBuilder_x8632raw::jumpIfNotEqual(std::string label) {
	jumpWithShortVersion(label, { 0x75 }, { 0x0f, 0x85 });
}

void AsmBuilder_x8632raw::jumpIfGreater(std::string label) {
	jumpWithShortVersion(label, { 0x7f }, { 0x0f, 0x8f });
}

void AsmBuilder_x8632raw::jumpIfGreaterOrEqual(std::string label) {
	jumpWithShortVersion(label, { 0x7d }, { 0x0f, 0x8d });
}

void AsmBuilder_x8632raw::noop() {
	raw8(0x90);
}

/* */

void AsmBuilder_x8632raw::callByLabel(std::string label) {
	raw8(0xe8);
	curSection->MissingReferences.emplace_back(
		BackfillableReference{ getCurrentSectionOffset(), BRT_RELATIVE_32, label }
	);
	raw32(0x00000000);
}

void AsmBuilder_x8632raw::callFromRegister(Register r) {
	switch (r) {
		case REG_A:		return raw8s({ 0xff, 0xd0 });
		case REG_C:		return raw8s({ 0xff, 0xd1 });
		case REG_D:		return raw8s({ 0xff, 0xd2 });
		case REG_B:		return raw8s({ 0xff, 0xd3 });
		case REG_SP:	return raw8s({ 0xff, 0xd4 });
		case REG_BP:	return raw8s({ 0xff, 0xd5 });
		case REG_SI:	return raw8s({ 0xff, 0xd6 });
		case REG_DI:	return raw8s({ 0xff, 0xd7 });
	}
}

void AsmBuilder_x8632raw::leave() {
	raw8(0xc9);
}

void AsmBuilder_x8632raw::returnFunc() {
	raw8(0xc3);
}

void AsmBuilder_x8632raw::returnFunc(uint16_t stacksize) {
	raw8(0xc2);
	raw16(stacksize);
}

/* */

bool AsmBuilder_x8632raw::patchReferences() {

	bool everythingWentOK = true;

	for (auto& section : sections) {
		auto & missingrefs = section.second.MissingReferences;
		for (auto ref = missingrefs.end(); ref != missingrefs.begin(); --ref) { // Iterate backward for easy element removal

			auto ltpos = labels.find(ref->ref_target);
			if (ltpos == labels.end()) {
				errors << "FATAL: Couldn't resolve label '" << ref->ref_target << "'\n";
				everythingWentOK = false;
				continue;
			}

			uint32_t set_value = 0;

			uint32_t current_value = 0;
			memcpy(
				&current_value,
				/*ltpos->second.*/ section.second.data.c_str() + ref->datapos,
				sizeof(uint32_t)
				);

			set_value += current_value;

			switch (ref->type) {
				case BRT_RELATIVE_32: {
					// Difference between the target location and the referent's program counter
					set_value += (
						ltpos->second.offset - ref->datapos
					);
				} break;

				default: {
					errors << "FATAL: Unexpected backreference type\n";
					everythingWentOK = false;
					continue;
				} break;
			}

			section.second.data[ref->datapos + 0] = (set_value >> 0) & 0xFF;
			section.second.data[ref->datapos + 1] = (set_value >> 8) & 0xFF;
			section.second.data[ref->datapos + 2] = (set_value >> 16) & 0xFF;
			section.second.data[ref->datapos + 3] = (set_value >> 24) & 0xFF;

			// Remove reference so it doesn't get processed again in the future
			missingrefs.erase(ref);

		}

	}

	return everythingWentOK;
}

bool AsmBuilder_x8632raw::renderObject(std::string& ret) {

	ret.clear();

	// We're in raw mode, there are no sections.
	// Require all relocations to be present

	if (!patchReferences()) {
		return false;
	}

	// Then just join them all

	for(auto& section : sections) {
		ret += section.second.data;
	}

	return true;
}
